entry.preferences-search {
  border: none;
  border-right: 1px solid alpha(@borders, 0.55);
  border-bottom: 1px solid alpha(@borders, 0.55);
  border-radius: 0;
}

dzlpreferencesview stacksidebar list {
  background-color: @content_view_bg;
  border-right: 1px solid alpha(@borders, 0.55);
}

dzlpreferencesview stacksidebar list separator {
  background-color: transparent;
}

dzlpreferencesview list row entry {
  background: transparent;
  border: none;
  padding: 0;
  padding-left: 5px;
  border-radius: 3px;
  margin: -5px;
}

dzlpreferencesview list row spinbutton entry {
  margin-left: 2px;
}

dzlpreferencesview dzlpreferencesgroup list row {
  padding: 10px;
  border-bottom: 1px solid alpha(@borders, 0.2);
}

dzlpreferencesview dzlpreferencesgroup list row:last-child {
  border-bottom: none;
}

dzlpreferencesview dzlpreferencesgroup list entry {
  background: none;
  min-height: 0px;
}

dzlpreferencesview stackswitcher {
  margin: 12px 12px 0px 12px;
}
