/* dzl-preferences-file-chooser-button.h
 *
 * Copyright (C) 2016 Akshaya Kakkilaya <akshaya.kakkilaya@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DZL_PREFERENCES_FILE_CHOOSER_BUTTON_H
#define DZL_PREFERENCES_FILE_CHOOSER_BUTTON_H

#include "dzl-version-macros.h"

#include "prefs/dzl-preferences-bin.h"

G_BEGIN_DECLS

#define DZL_TYPE_PREFERENCES_FILE_CHOOSER_BUTTON (dzl_preferences_file_chooser_button_get_type())

DZL_AVAILABLE_IN_ALL
G_DECLARE_FINAL_TYPE (DzlPreferencesFileChooserButton, dzl_preferences_file_chooser_button, DZL, PREFERENCES_FILE_CHOOSER_BUTTON, DzlPreferencesBin)

G_END_DECLS

#endif /* DZL_PREFERENCES_FILE_CHOOSER_BUTTON_H */
