dzlsuggestionpopover {
  background: transparent;
}

dzlsuggestionpopover > revealer > box {
  margin: 5px 6px 6px 6px;
  border: 1px solid @borders;
  box-shadow: 0px 0px 3px @wm_shadow;
  padding: 3px 3px 0px 3px;
  background: @content_view_bg;
  border-radius: 5px;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow {
  background: alpha(@content_view_bg, 0.97);
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list {
  background: transparent;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row:first-child {
  transition: none;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row {
  border-radius: 3px;
  transition: none;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row:selected .title {
  color: @theme_selected_fg_color;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row .title {
  margin-left: 12px;
  color: @theme_fg_color;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row .separator {
  margin-left: 12px;
  margin-right: 12px;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row > box {
  margin: 4px 8px;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row:last-child {
  border-bottom: none;
  margin-bottom: 3px;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row > box > image:first-child {
  min-width: 16px;
}
