button.dzlmenubutton > box {
  margin-left: 3px;
  margin-right: 3px;
}
button.dzlmenubutton > box > image.arrow {
  margin-left: 3px;
}
popover.dzlmenubutton > box {
  margin: 15px 15px 15px 12px;
}
dzlmenubuttonsection:first-child separator {
  background: none;
  margin-top: 0px;
  margin-bottom: 0px;
}
dzlmenubuttonsection:not(:first-child) separator {
  border-color: @borders;
  margin-top: 8px;
  margin-bottom: 8px;
  margin-left: 10px;
  margin-right: 10px;
}
dzlmenubuttonsection > label.title {
  font-weight: bold;
  font-size: 0.8333em;
  opacity: 0.5;
  margin-bottom: 6px;
}
dzlmenubuttonsection box.vertical checkbutton.dzlmenubuttonitem,
dzlmenubuttonsection box.vertical button.dzlmenubuttonitem {
  border: none;
  background: none;
  box-shadow: none;
  padding: 1px 10px;
  outline-offset: -2px;
}
dzlmenubuttonsection box.vertical checkbutton.dzlmenubuttonitem:hover,
dzlmenubuttonsection box.vertical button.dzlmenubuttonitem:hover {
  background-color: alpha(@theme_fg_color, 0.05);
}
dzlmenubuttonsection box.vertical checkbutton.dzlmenubuttonitem > box > image:first-child,
dzlmenubuttonsection box.vertical button.dzlmenubuttonitem > box > image:first-child {
  margin: 3px 8px 3px 0px;
  color: @theme_fg_color;
}
