dzlsearchbar box.search-bar {
  background-color: #d6d6d6;
  box-shadow: 0 4px 3px -5px #aaa inset,
              0 -1px 0 #a1a1a1 inset;
}
dzlsearchbar:backdrop box.search-bar {
  background-color: #d5d5d5;
  box-shadow: none;
}
